from unittest import IsolatedAsyncioTestCase

from .....admin.request_context import AdminRequestContext
from .....storage.error import StorageError
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import routes as test_module
from ..manager import V10DiscoveryMgr
from ..messages.query import Query
from ..models.discovery_record import V10DiscoveryExchangeRecord


class TestDiscoveryRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_query_features(self):
        self.request.json = mock.CoroutineMock()

        self.request.query = {"query": "*"}

        test_rec = V10DiscoveryExchangeRecord(
            discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
            query_msg=Query(query="*"),
        )
        with (
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(
                V10DiscoveryMgr, "create_and_send_query", autospec=True
            ) as mock_create_query,
        ):
            mock_create_query.return_value = test_rec
            await test_module.query_features(self.request)
            mock_response.assert_called_once_with(test_rec.serialize())

    async def test_query_features_with_connection(self):
        self.request.json = mock.CoroutineMock()

        self.request.query = {"query": "*", "connection_id": "test", "comment": "test"}

        test_rec = V10DiscoveryExchangeRecord(
            discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
            query_msg=Query(query="*"),
        )

        with (
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(
                V10DiscoveryMgr, "create_and_send_query", autospec=True
            ) as mock_create_query,
        ):
            mock_create_query.return_value = test_rec
            await test_module.query_features(self.request)
            mock_response.assert_called_once_with(test_rec.serialize())

    async def test_query_records(self):
        self.request.json = mock.CoroutineMock()

        self.request.query = {"connection_id": "test"}

        test_rec = V10DiscoveryExchangeRecord(
            discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
            query_msg=Query(query="*"),
        )

        with (
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(
                test_module, "V10DiscoveryExchangeRecord", autospec=True
            ) as mock_ex_rec,
        ):
            mock_ex_rec.retrieve_by_connection_id.return_value = test_rec
            await test_module.query_records(self.request)
            mock_response.assert_called_once_with({"results": [test_rec.serialize()]})

    async def test_query_records_x(self):
        self.request.json = mock.CoroutineMock()

        self.request.query = {"connection_id": "test"}

        with (
            mock.patch.object(test_module.web, "json_response"),
            mock.patch.object(
                test_module, "V10DiscoveryExchangeRecord", autospec=True
            ) as mock_ex_rec,
        ):
            mock_ex_rec.retrieve_by_connection_id.side_effect = StorageError
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.query_records(self.request)

    async def test_query_records_all(self):
        self.request.json = mock.CoroutineMock()

        test_recs = [
            V10DiscoveryExchangeRecord(
                discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
                query_msg=Query(query="*"),
            ),
            V10DiscoveryExchangeRecord(
                discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa7",
                query_msg=Query(query="test"),
            ),
        ]

        with (
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(
                test_module, "V10DiscoveryExchangeRecord", autospec=True
            ) as mock_ex_rec,
        ):
            mock_ex_rec.query.return_value = test_recs
            await test_module.query_records(self.request)
            mock_response.assert_called_once_with(
                {"results": [k.serialize() for k in test_recs]}
            )

    async def test_query_records_connection_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module.web, "json_response"),
            mock.patch.object(
                test_module, "V10DiscoveryExchangeRecord", autospec=True
            ) as mock_ex_rec,
        ):
            mock_ex_rec.query.side_effect = StorageError
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.query_records(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



DOC = {
    "@context": "https://w3id.org/did/v1",
    "id": "did:example:1234abcd",
    "verificationMethod": [
        {
            "id": "did:example:1234abcd#4",
            "type": "RsaVerificationKey2018",
            "controller": "did:example:1234abcd",
            "publicKeyPem": "-----BEGIN PUBLIC X…",
        },
        {
            "id": "did:example:1234abcd#5",
            "type": "RsaVerificationKey2018",
            "controller": "did:example:1234abcd",
            "publicKeyPem": "-----BEGIN PUBLIC 9…",
        },
        {
            "id": "did:example:1234abcd#6",
            "type": "RsaVerificationKey2018",
            "controller": "did:example:1234abcd",
            "publicKeyPem": "-----BEGIN PUBLIC A…",
        },
    ],
    "authentication": [
        {
            "id": "did:example:123456789abcdefghi#ted",
            "controller": "did:example:1234abcd",
            "type": "RsaSignatureAuthentication2018",
            "publicKey": "did:example:1234abcd#4",
        },
        "did:example:123456789abcdefghi#5",
    ],
    "service": [
        {
            "id": "did:example:123456789abcdefghi#did-communication",
            "type": "did-communication",
            "priority": 0,
            "recipientKeys": ["did:example:1234abcd#4"],
            "routingKeys": ["did:example:1234abcd#6"],
            "serviceEndpoint": "http://example.com",
        }
    ],
}



"""Rotate handler."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from ..manager import DIDRotateManager
from ..messages.rotate import Rotate


class RotateHandler(BaseHandler):
    """Message handler class for rotate message."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle rotate message.

        Args:
            context: request context
            responder: responder callback
        """
        self._logger.debug("RotateHandler called with context %s", context)
        assert isinstance(context.message, Rotate)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        connection_record = context.connection_record
        rotate = context.message

        profile = context.profile
        did_rotate_mgr = DIDRotateManager(profile)

        if record := await did_rotate_mgr.receive_rotate(connection_record, rotate):
            await did_rotate_mgr.commit_rotate(connection_record, record)



from .bbs_bls_signature_2020 import BbsBlsSignature2020 as _BbsBlsSignature2020
from .bbs_bls_signature_proof_2020 import (
    BbsBlsSignatureProof2020 as _BbsBlsSignatureProof2020,
)
from .ecdsa_secp256r1_signature_2019 import (
    EcdsaSecp256r1Signature2019 as _EcdsaSecp256r1Signature2019,
)
from .ed25519_signature_2018 import Ed25519Signature2018 as _Ed25519Signature2018
from .ed25519_signature_2020 import Ed25519Signature2020 as _Ed25519Signature2020
from .jws_linked_data_signature import JwsLinkedDataSignature as _JwsLinkedDataSignature
from .linked_data_proof import LinkedDataProof as _LinkedDataProof
from .linked_data_signature import LinkedDataSignature as _LinkedDataSignature

__all__ = [
    "_LinkedDataProof",
    "_LinkedDataSignature",
    "_JwsLinkedDataSignature",
    "_Ed25519Signature2018",
    "_Ed25519Signature2020",
    "_EcdsaSecp256r1Signature2019",
    "_BbsBlsSignature2020",
    "_BbsBlsSignatureProof2020",
]



"""Multiple IndyVdrLedger Manager."""

import asyncio
import concurrent.futures
import json
import logging
from collections import OrderedDict
from typing import List, Mapping, Optional, Tuple

from ...cache.base import BaseCache
from ...core.profile import Profile
from ...ledger.base import BaseLedger
from ...ledger.error import LedgerError
from ...wallet.crypto import did_is_self_certified
from ..indy_vdr import IndyVdrLedger
from ..merkel_validation.domain_txn_handler import get_proof_nodes, prepare_for_state_read
from ..merkel_validation.trie import SubTrie
from .base_manager import BaseMultipleLedgerManager, MultipleLedgerManagerError

LOGGER = logging.getLogger(__name__)


class MultiIndyVDRLedgerManager(BaseMultipleLedgerManager):
    """Multiple Indy VDR Ledger Manager."""

    def __init__(
        self,
        profile: Profile,
        production_ledgers: Optional[OrderedDict] = None,
        non_production_ledgers: Optional[OrderedDict] = None,
        writable_ledgers: Optional[set] = None,
        endorser_map: Optional[dict] = None,
        cache_ttl: Optional[int] = None,
    ):
        """Initialize MultiIndyLedgerManager.

        Args:
            profile (Profile): The base profile for this manager.
            production_ledgers (Optional[OrderedDict]): Mapping of production
                IndyVDRLedger.
            non_production_ledgers (Optional[OrderedDict]): Mapping of non-production
                IndyVDRLedger.
            writable_ledgers (Optional[set]): Set of writable ledgers.
            endorser_map (Optional[dict]): Mapping of endorsers.
            cache_ttl (int): Time in seconds to persist did_ledger_id_resolver cache keys.

        """
        self.profile = profile
        self.production_ledgers = production_ledgers or OrderedDict()
        self.non_production_ledgers = non_production_ledgers or OrderedDict()
        self.writable_ledgers = writable_ledgers or set()
        self.endorser_map = endorser_map or {}
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=5)
        self.cache_ttl = cache_ttl

    async def get_write_ledgers(self) -> List[str]:
        """Return the write IndyVdrLedger instance."""
        return list(self.writable_ledgers)

    def get_endorser_info_for_ledger(self, ledger_id: str) -> Optional[Tuple[str, str]]:
        """Return endorser alias, did tuple for provided ledger, if available."""
        endorser_info = self.endorser_map.get(ledger_id)
        if not endorser_info:
            return None
        return (endorser_info["endorser_alias"], endorser_info["endorser_did"])

    async def get_prod_ledgers(self) -> Mapping:
        """Return production ledgers mapping."""
        return self.production_ledgers

    async def get_nonprod_ledgers(self) -> Mapping:
        """Return non_production ledgers mapping."""
        return self.non_production_ledgers

    async def get_ledger_inst_by_id(self, ledger_id: str) -> Optional[BaseLedger]:
        """Return BaseLedger instance."""
        return self.production_ledgers.get(ledger_id) or self.non_production_ledgers.get(
            ledger_id
        )

    async def get_ledger_id_by_ledger_pool_name(self, pool_name: str) -> str:
        """Return ledger_id by ledger pool name."""
        multi_ledgers = self.production_ledgers | self.non_production_ledgers
        for ledger_id, indy_vdr_ledger in multi_ledgers.items():
            if indy_vdr_ledger.pool_name == pool_name:
                return ledger_id
        raise MultipleLedgerManagerError(
            f"Provided Ledger pool name {pool_name} not found "
            "in either production_ledgers or non_production_ledgers"
        )

    async def _get_ledger_by_did(
        self,
        ledger_id: str,
        did: str,
    ) -> Optional[Tuple[str, IndyVdrLedger, bool]]:
        """Build and submit GET_NYM request and process response.

        Successful response return tuple with ledger_id, IndyVdrLedger instance
        and is_self_certified bool flag. Unsuccessful response return None.

        Args:
            ledger_id: provided ledger_id to retrieve IndyVdrLedger instance
                        from production_ledgers or non_production_ledgers
            did: provided DID

        Return:
            (str, IndyVdrLedger, bool) or None
        """
        try:
            indy_vdr_ledger = None
            if ledger_id in self.production_ledgers:
                indy_vdr_ledger = self.production_ledgers.get(ledger_id)
            else:
                indy_vdr_ledger = self.non_production_ledgers.get(ledger_id)
            async with indy_vdr_ledger:
                request = await indy_vdr_ledger.build_and_return_get_nym_request(
                    None, did
                )
                response_json = await asyncio.wait_for(
                    indy_vdr_ledger.submit_get_nym_request(request), 10
                )
                if isinstance(response_json, dict):
                    response = response_json
                else:
                    response = json.loads(response_json)
                if "result" in response.keys():
                    data = response.get("result", {}).get("data")
                else:
                    data = response.get("data")
                if not data:
                    LOGGER.warning(f"Did {did} not posted to ledger {ledger_id}")
                    return None
                if isinstance(data, str):
                    data = json.loads(data)
                if not await SubTrie.verify_spv_proof(
                    expected_value=prepare_for_state_read(response),
                    proof_nodes=get_proof_nodes(response),
                ):
                    LOGGER.warning(
                        f"State Proof validation failed for Did {did} "
                        f"and ledger {ledger_id}"
                    )
                    return None
                if did_is_self_certified(did, data.get("verkey")):
                    return (ledger_id, indy_vdr_ledger, True)
                return (ledger_id, indy_vdr_ledger, False)
        except asyncio.TimeoutError:
            LOGGER.exception(
                f"get-nym request timedout for Did {did} and "
                f"ledger {ledger_id}, reply not received within 10 sec"
            )
            return None
        except LedgerError as err:
            LOGGER.error(
                "Exception when building and submitting get-nym request, "
                f"for Did {did} and ledger {ledger_id}, {err}"
            )
            return None

    async def lookup_did_in_configured_ledgers(
        self, did: str, cache_did: bool = True
    ) -> Tuple[str, IndyVdrLedger]:
        """Lookup given DID in configured ledgers in parallel."""
        self.cache = self.profile.inject_or(BaseCache)
        cache_key = f"did_ledger_id_resolver::{did}"
        if bool(cache_did and self.cache and await self.cache.get(cache_key)):
            cached_ledger_id = await self.cache.get(cache_key)
            if cached_ledger_id in self.production_ledgers:
                return (cached_ledger_id, self.production_ledgers.get(cached_ledger_id))
            elif cached_ledger_id in self.non_production_ledgers:
                return (
                    cached_ledger_id,
                    self.non_production_ledgers.get(cached_ledger_id),
                )
            else:
                raise MultipleLedgerManagerError(
                    f"cached ledger_id {cached_ledger_id} not found in either "
                    "production_ledgers or non_production_ledgers"
                )
        applicable_prod_ledgers = {"self_certified": {}, "non_self_certified": {}}
        applicable_non_prod_ledgers = {"self_certified": {}, "non_self_certified": {}}
        ledger_ids = list(self.production_ledgers.keys()) + list(
            self.non_production_ledgers.keys()
        )
        coro_futures = {
            self.executor.submit(self._get_ledger_by_did, ledger_id, did): ledger_id
            for ledger_id in ledger_ids
        }
        for coro_future in concurrent.futures.as_completed(coro_futures):
            result = await coro_future.result()
            if result:
                applicable_ledger_id = result[0]
                applicable_ledger_inst = result[1]
                is_self_certified = result[2]
                if applicable_ledger_id in self.production_ledgers:
                    insert_key = list(self.production_ledgers).index(applicable_ledger_id)
                    if is_self_certified:
                        applicable_prod_ledgers["self_certified"][insert_key] = (
                            applicable_ledger_id,
                            applicable_ledger_inst,
                        )
                    else:
                        applicable_prod_ledgers["non_self_certified"][insert_key] = (
                            applicable_ledger_id,
                            applicable_ledger_inst,
                        )
                else:
                    insert_key = list(self.non_production_ledgers).index(
                        applicable_ledger_id
                    )
                    if is_self_certified:
                        applicable_non_prod_ledgers["self_certified"][insert_key] = (
                            applicable_ledger_id,
                            applicable_ledger_inst,
                        )
                    else:
                        applicable_non_prod_ledgers["non_self_certified"][insert_key] = (
                            applicable_ledger_id,
                            applicable_ledger_inst,
                        )
        applicable_prod_ledgers["self_certified"] = OrderedDict(
            sorted(applicable_prod_ledgers.get("self_certified").items())
        )
        applicable_prod_ledgers["non_self_certified"] = OrderedDict(
            sorted(applicable_prod_ledgers.get("non_self_certified").items())
        )
        applicable_non_prod_ledgers["self_certified"] = OrderedDict(
            sorted(applicable_non_prod_ledgers.get("self_certified").items())
        )
        applicable_non_prod_ledgers["non_self_certified"] = OrderedDict(
            sorted(applicable_non_prod_ledgers.get("non_self_certified").items())
        )
        if len(applicable_prod_ledgers.get("self_certified")) > 0:
            successful_ledger_inst = list(
                applicable_prod_ledgers.get("self_certified").values()
            )[0]
            if cache_did and self.cache:
                await self.cache.set(cache_key, successful_ledger_inst[0], self.cache_ttl)
            return successful_ledger_inst
        elif len(applicable_non_prod_ledgers.get("self_certified")) > 0:
            successful_ledger_inst = list(
                applicable_non_prod_ledgers.get("self_certified").values()
            )[0]
            if cache_did and self.cache:
                await self.cache.set(cache_key, successful_ledger_inst[0], self.cache_ttl)
            return successful_ledger_inst
        elif len(applicable_prod_ledgers.get("non_self_certified")) > 0:
            successful_ledger_inst = list(
                applicable_prod_ledgers.get("non_self_certified").values()
            )[0]
            if cache_did and self.cache:
                await self.cache.set(cache_key, successful_ledger_inst[0], self.cache_ttl)
            return successful_ledger_inst
        elif len(applicable_non_prod_ledgers.get("non_self_certified")) > 0:
            successful_ledger_inst = list(
                applicable_non_prod_ledgers.get("non_self_certified").values()
            )[0]
            if cache_did and self.cache:
                await self.cache.set(cache_key, successful_ledger_inst[0], self.cache_ttl)
            return successful_ledger_inst
        else:
            raise MultipleLedgerManagerError(
                f"DID {did} not found in any of the ledgers total: "
                f"(production: {len(self.production_ledgers)}, "
                f"non_production: {len(self.non_production_ledgers)})"
            )



from unittest import IsolatedAsyncioTestCase

import pytest

from .....messaging.responder import BaseResponder, MockResponder
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from ....didcomm_prefix import DIDCommPrefix
from ..manager import V10DiscoveryMgr
from ..messages.disclose import Disclose
from ..messages.query import Query
from ..models.discovery_record import V10DiscoveryExchangeRecord

TEST_DISCOVERY_EX_REC = V10DiscoveryExchangeRecord(
    discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6"
)


class TestV10DiscoveryManager(IsolatedAsyncioTestCase):
    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.context = self.profile.context
        self.manager = V10DiscoveryMgr(self.profile)
        self.disclose = Disclose(
            protocols=[
                {
                    "pid": DIDCommPrefix.qualify_current("test_proto/v1.0/test_message"),
                    "roles": [],
                }
            ]
        )
        self.query = Query(query="*")
        self.query.assign_thread_id("test123")
        self.responder = MockResponder()
        self.profile.context.injector.bind_instance(BaseResponder, self.responder)
        assert self.manager.profile

    async def test_receive_disclosure(self):
        test_conn_id = "test123"
        self.disclose.assign_thread_id("test123")
        with (
            mock.patch.object(
                V10DiscoveryExchangeRecord, "save", autospec=True
            ) as save_ex,
            mock.patch.object(
                V10DiscoveryExchangeRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve,
        ):
            mock_retrieve.return_value = TEST_DISCOVERY_EX_REC
            ex_rec = await self.manager.receive_disclose(
                disclose_msg=self.disclose, connection_id=test_conn_id
            )
            save_ex.assert_called_once()
            assert ex_rec == V10DiscoveryExchangeRecord(
                discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
                disclose=self.disclose,
                connection_id=test_conn_id,
            )

    async def test_receive_disclosure_retrieve_by_conn(self):
        test_conn_id = "test123"
        self.disclose.assign_thread_id("test123")
        with (
            mock.patch.object(
                V10DiscoveryExchangeRecord, "save", autospec=True
            ) as save_ex,
            mock.patch.object(
                V10DiscoveryExchangeRecord,
                "exists_for_connection_id",
                mock.CoroutineMock(),
            ) as mock_exists_for_connection_id,
            mock.patch.object(
                V10DiscoveryExchangeRecord,
                "retrieve_by_connection_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_by_connection_id,
            mock.patch.object(
                V10DiscoveryExchangeRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve_by_id,
        ):
            mock_retrieve_by_id.side_effect = StorageNotFoundError
            mock_exists_for_connection_id.return_value = True
            mock_retrieve_by_connection_id.return_value = TEST_DISCOVERY_EX_REC
            ex_rec = await self.manager.receive_disclose(
                disclose_msg=self.disclose, connection_id=test_conn_id
            )
            save_ex.assert_called_once()
            assert ex_rec == V10DiscoveryExchangeRecord(
                discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
                disclose=self.disclose,
                connection_id=test_conn_id,
            )

    async def test_receive_disclosure_retrieve_by_conn_not_found(self):
        test_conn_id = "test123"
        self.disclose.assign_thread_id("test123")
        with (
            mock.patch.object(
                V10DiscoveryExchangeRecord, "save", autospec=True
            ) as save_ex,
            mock.patch.object(
                V10DiscoveryExchangeRecord,
                "exists_for_connection_id",
                mock.CoroutineMock(),
            ) as mock_exists_for_connection_id,
            mock.patch.object(
                V10DiscoveryExchangeRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve_by_id,
        ):
            mock_retrieve_by_id.side_effect = StorageNotFoundError
            mock_exists_for_connection_id.return_value = False
            ex_rec = await self.manager.receive_disclose(
                disclose_msg=self.disclose, connection_id=test_conn_id
            )
            save_ex.assert_called_once()
            assert ex_rec == V10DiscoveryExchangeRecord(
                discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
                disclose=self.disclose,
                connection_id=test_conn_id,
            )

    async def test_receive_disclosure_retrieve_new_ex_rec(self):
        test_conn_id = "test123"
        with (
            mock.patch.object(
                V10DiscoveryExchangeRecord, "save", autospec=True
            ) as save_ex,
            mock.patch.object(
                V10DiscoveryExchangeRecord,
                "exists_for_connection_id",
                mock.CoroutineMock(),
            ) as mock_exists_for_connection_id,
            mock.patch.object(
                V10DiscoveryExchangeRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve_by_id,
        ):
            mock_retrieve_by_id.side_effect = StorageNotFoundError
            mock_exists_for_connection_id.return_value = False
            ex_rec = await self.manager.receive_disclose(
                disclose_msg=self.disclose, connection_id=test_conn_id
            )
            save_ex.assert_called_once()
            assert ex_rec == V10DiscoveryExchangeRecord(
                disclose=self.disclose,
                connection_id=test_conn_id,
            )

    async def test_check_if_disclosure_received(self):
        with mock.patch.object(
            V10DiscoveryExchangeRecord,
            "retrieve_by_id",
            mock.CoroutineMock(),
        ) as mock_retrieve_by_id:
            mock_retrieve_by_id.side_effect = [
                V10DiscoveryExchangeRecord(),
                V10DiscoveryExchangeRecord(disclose=Disclose()),
            ]
            assert await self.manager.check_if_disclosure_received(record_id="test123")

    async def test_create_and_send_query_with_connection(self):
        return_ex_rec = V10DiscoveryExchangeRecord(query_msg=Query(query="*"))
        with (
            mock.patch.object(
                V10DiscoveryExchangeRecord,
                "exists_for_connection_id",
                mock.CoroutineMock(),
            ) as mock_exists_for_connection_id,
            mock.patch.object(
                V10DiscoveryExchangeRecord,
                "retrieve_by_connection_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_by_connection_id,
            mock.patch.object(
                V10DiscoveryExchangeRecord,
                "save",
                mock.CoroutineMock(),
            ),
            mock.patch.object(
                V10DiscoveryMgr, "check_if_disclosure_received", mock.CoroutineMock()
            ) as mock_disclosure_received,
            mock.patch.object(self.responder, "send", mock.CoroutineMock()) as mock_send,
        ):
            mock_exists_for_connection_id.return_value = True
            mock_retrieve_by_connection_id.return_value = V10DiscoveryExchangeRecord()
            mock_disclosure_received.return_value = return_ex_rec
            received_ex_rec = await self.manager.create_and_send_query(
                query="*", connection_id="test123"
            )
            assert received_ex_rec.query_msg == return_ex_rec.query_msg
            mock_send.assert_called_once()

    async def test_create_and_send_query_with_no_connection(self):
        with mock.patch.object(
            V10DiscoveryMgr,
            "receive_query",
            mock.CoroutineMock(),
        ) as mock_receive_query:
            mock_receive_query.return_value = Disclose()
            received_ex_rec = await self.manager.create_and_send_query(query="*")
            assert received_ex_rec.query_msg.query == "*"
            assert received_ex_rec.disclose.protocols == []



"""Handler for mediate-request message."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from ..manager import MediationAlreadyExists, MediationManager
from ..messages.mediate_request import MediationRequest
from ..messages.problem_report import CMProblemReport, ProblemReportReason


class MediationRequestHandler(BaseHandler):
    """Handler for mediate-request message."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle mediate-request message."""
        self._logger.debug("%s called with context %s", self.__class__.__name__, context)
        assert isinstance(context.message, MediationRequest)

        if not context.connection_ready:
            raise HandlerException("Invalid mediation request: no active connection")

        mgr = MediationManager(context.profile)
        try:
            record = await mgr.receive_request(
                context.connection_record.connection_id, context.message
            )
            if context.settings.get("mediation.open", False):
                record, grant = await mgr.grant_request(record.mediation_id)
                await responder.send_reply(grant)
        except MediationAlreadyExists:
            reply = CMProblemReport(
                description={
                    "en": "Mediation request already exists from this connection",
                    "code": ProblemReportReason.MEDIATION_REQUEST_REPEAT.value,
                }
            )
            reply.assign_thread_from(context.message)
            await responder.send_reply(reply)



from .prove import (
    _load_w3c_credentials,
    create_signed_anoncreds_presentation,
    prepare_data_for_presentation,
)
from .verify import verify_signed_anoncredspresentation

__all__ = [
    "verify_signed_anoncredspresentation",
    "create_signed_anoncreds_presentation",
    "prepare_data_for_presentation",
    "_load_w3c_credentials",
]



import json
from time import time
from unittest import IsolatedAsyncioTestCase

from .....indy.holder import IndyHolder, IndyHolderError
from .....indy.models.pres_preview import (
    IndyPresAttrSpec,
    IndyPresPredSpec,
    IndyPresPreview,
)
from .....indy.models.xform import indy_proof_req_preview2indy_requested_creds
from .....indy.verifier import IndyVerifier
from .....ledger.base import BaseLedger
from .....ledger.multiple_ledger.ledger_requests_executor import (
    IndyLedgerRequestsExecutor,
)
from .....messaging.decorators.attach_decorator import AttachDecorator
from .....messaging.responder import BaseResponder, MockResponder
from .....protocols.issue_credential.v1_0.models.credential_exchange import (
    V10CredentialExchange,
)
from .....tests import mock
from .....utils.testing import create_test_profile
from ....didcomm_prefix import DIDCommPrefix
from ...indy import pres_exch_handler as test_indy_util_module
from .. import manager as test_module
from ..manager import PresentationManager, PresentationManagerError
from ..message_types import ATTACH_DECO_IDS, PRESENTATION, PRESENTATION_REQUEST
from ..messages.presentation import Presentation
from ..messages.presentation_problem_report import PresentationProblemReport
from ..messages.presentation_proposal import PresentationProposal
from ..messages.presentation_request import PresentationRequest
from ..models.presentation_exchange import V10PresentationExchange

NOW = int(time())
CONN_ID = "connection_id"
ISSUER_DID = "NcYxiDXkpYi6ov5FcYDi1e"
S_ID = f"{ISSUER_DID}:2:vidya:1.0"
CD_ID = f"{ISSUER_DID}:3:CL:{S_ID}:tag1"
RR_ID = f"{ISSUER_DID}:4:{CD_ID}:CL_ACCUM:0"
PRES_PREVIEW = IndyPresPreview(
    attributes=[
        IndyPresAttrSpec(name="player", cred_def_id=CD_ID, value="Richie Knucklez"),
        IndyPresAttrSpec(
            name="screenCapture",
            cred_def_id=CD_ID,
            mime_type="image/png",
            value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
        ),
    ],
    predicates=[
        IndyPresPredSpec(
            name="highScore", cred_def_id=CD_ID, predicate=">=", threshold=1000000
        )
    ],
)
PRES_PREVIEW_NAMES = IndyPresPreview(
    attributes=[
        IndyPresAttrSpec(
            name="player", cred_def_id=CD_ID, value="Richie Knucklez", referent="0"
        ),
        IndyPresAttrSpec(
            name="screenCapture",
            cred_def_id=CD_ID,
            mime_type="image/png",
            value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
            referent="0",
        ),
    ],
    predicates=[
        IndyPresPredSpec(
            name="highScore", cred_def_id=CD_ID, predicate=">=", threshold=1000000
        )
    ],
)
PROOF_REQ_NAME = "name"
PROOF_REQ_VERSION = "1.0"
PROOF_REQ_NONCE = "12345"
INDY_PROOF = {
    "proof": {
        "proofs": [
            {
                "primary_proof": {
                    "eq_proof": {
                        "revealed_attrs": {
                            "player": "51643998292319337989",
                            "screencapture": "124831723185628395682368329568235681",
                        },
                        "a_prime": "98381845469564775640588",
                        "e": "2889201651469315129053056279820725958192110265136",
                        "v": "337782521199137176224",
                        "m": {
                            "master_secret": "88675074759262558623",
                            "date": "3707627155679953691027082306",
                            "highscore": "251972383037120760793174059437326",
                        },
                        "m2": "2892781443118611948331343540849982215419978654911341",
                    },
                    "ge_proofs": [
                        {
                            "u": {
                                "0": "99189584890680947709857922351898933228959",
                                "3": "974568160016086782335901983921278203",
                                "2": "127290395299",
                                "1": "7521808223922",
                            },
                            "r": {
                                "3": "247458",
                                "2": "263046",
                                "1": "285214",
                                "DELTA": "4007402",
                                "0": "12066738",
                            },
                            "mj": "1507606",
                            "alpha": "20251550018805200",
                            "t": {
                                "1": "1262519732727",
                                "3": "82102416",
                                "0": "100578099981822",
                                "2": "47291",
                                "DELTA": "556736142765",
                            },
                            "predicate": {
                                "attr_name": "highscore",
                                "p_type": "GE",
                                "value": 1000000,
                            },
                        }
                    ],
                },
                "non_revoc_proof": {
                    "x_list": {
                        "rho": "128121489ACD4D778ECE",
                        "r": "1890DEFBB8A254",
                        "r_prime": "0A0861FFE96C",
                        "r_prime_prime": "058376CE",
                        "r_prime_prime_prime": "188DF30745A595",
                        "o": "0D0F7FA1",
                        "o_prime": "28165",
                        "m": "0187A9817897FC",
                        "m_prime": "91261D96B",
                        "t": "10FE96",
                        "t_prime": "10856A",
                        "m2": "B136089AAF",
                        "s": "018969A6D",
                        "c": "09186B6A",
                    },
                    "c_list": {
                        "e": "6 1B161",
                        "d": "6 19E861869",
                        "a": "6 541441EE2",
                        "g": "6 7601B068C",
                        "w": "21 10DE6 4 AAAA 5 2458 6 16161",
                        "s": "21 09616 4 1986 5 9797 6 BBBBB",
                        "u": "21 3213123 4 0616FFE 5 323 6 110861861",
                    },
                },
            }
        ],
        "aggregated_proof": {
            "c_hash": "81147637626525127013830996",
            "c_list": [
                [3, 18, 46, 12],
                [3, 136, 2, 39],
                [100, 111, 148, 193],
                [1, 123, 11, 152],
                [2, 138, 162, 227],
                [1, 239, 33, 47],
            ],
        },
    },
    "requested_proof": {
        "revealed_attrs": {
            "0_player_uuid": {
                "sub_proof_index": 0,
                "raw": "Richie Knucklez",
                "encoded": "516439982",
            },
            "0_screencapture_uuid": {
                "sub_proof_index": 0,
                "raw": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                "encoded": "4434954949",
            },
        },
        "self_attested_attrs": {},
        "unrevealed_attrs": {},
        "predicates": {"0_highscore_GE_uuid": {"sub_proof_index": 0}},
    },
    "identifiers": [
        {
            "schema_id": S_ID,
            "cred_def_id": CD_ID,
            "rev_reg_id": RR_ID,
            "timestamp": NOW,
        }
    ],
}
PRES = Presentation(
    comment="Test",
    presentations_attach=[
        AttachDecorator.data_base64(
            mapping=INDY_PROOF,
            ident=ATTACH_DECO_IDS[PRESENTATION],
        )
    ],
)
PRES.assign_thread_id("dummy")


class TestPresentationManager(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        injector = self.profile.context.injector

        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=mock.MagicMock())
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {"revocation": {"...": "..."}}}
        )
        self.ledger.get_revoc_reg_def = mock.CoroutineMock(
            return_value={
                "ver": "1.0",
                "id": RR_ID,
                "revocDefType": "CL_ACCUM",
                "tag": RR_ID.split(":")[-1],
                "credDefId": CD_ID,
                "value": {
                    "IssuanceType": "ISSUANCE_BY_DEFAULT",
                    "maxCredNum": 1000,
                    "publicKeys": {"accumKey": {"z": "1 ..."}},
                    "tailsHash": "3MLjUFQz9x9n5u9rFu8Ba9C5bo4HNFjkPNc54jZPSNaZ",
                    "tailsLocation": "http://sample.ca/path",
                },
            }
        )
        self.ledger.get_revoc_reg_delta = mock.CoroutineMock(
            return_value=(
                {
                    "ver": "1.0",
                    "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1]},
                },
                NOW,
            )
        )
        self.ledger.get_revoc_reg_entry = mock.CoroutineMock(
            return_value=(
                {
                    "ver": "1.0",
                    "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1]},
                },
                NOW,
            )
        )
        injector.bind_instance(BaseLedger, self.ledger)
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_did = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        injector.bind_instance(IndyLedgerRequestsExecutor, mock_executor)
        self.holder = mock.MagicMock(IndyHolder, autospec=True)
        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {
                        "referent": "dummy_reft",
                        "attrs": {
                            "player": "Richie Knucklez",
                            "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "highScore": "1234560",
                        },
                    }
                },  # leave this comma: return a tuple
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds
        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "schema_id": S_ID,
                    "cred_def_id": CD_ID,
                    "rev_reg_id": RR_ID,
                    "cred_rev_id": 1,
                }
            )
        )
        self.holder.create_presentation = mock.CoroutineMock(return_value="{}")
        self.holder.create_revocation_state = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "witness": {"omega": "1 ..."},
                    "rev_reg": {"accum": "21 ..."},
                    "timestamp": NOW,
                }
            )
        )
        injector.bind_instance(IndyHolder, self.holder)

        self.verifier = mock.MagicMock(IndyVerifier, autospec=True)
        self.verifier.verify_presentation = mock.CoroutineMock(return_value=("true", []))
        injector.bind_instance(IndyVerifier, self.verifier)

        self.manager = PresentationManager(self.profile)

    async def test_record_eq(self):
        same = [
            V10PresentationExchange(
                presentation_exchange_id="dummy-0",
                thread_id="thread-0",
                role=V10PresentationExchange.ROLE_PROVER,
            )
        ] * 2
        diff = [
            V10PresentationExchange(
                presentation_exchange_id="dummy-1",
                role=V10PresentationExchange.ROLE_PROVER,
            ),
            V10PresentationExchange(
                presentation_exchange_id="dummy-0",
                thread_id="thread-1",
                role=V10PresentationExchange.ROLE_PROVER,
            ),
            V10PresentationExchange(
                presentation_exchange_id="dummy-1",
                thread_id="thread-0",
                role=V10PresentationExchange.ROLE_VERIFIER,
            ),
        ]

        for i in range(len(same) - 1):
            for j in range(i, len(same)):
                assert same[i] == same[j]

        for i in range(len(diff) - 1):
            for j in range(i, len(diff)):
                assert diff[i] == diff[j] if i == j else diff[i] != diff[j]

    async def test_create_exchange_for_proposal(self):
        proposal = PresentationProposal()

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(PresentationProposal, "serialize", autospec=True),
        ):
            exchange = await self.manager.create_exchange_for_proposal(
                CONN_ID,
                proposal,
                auto_present=None,
                auto_remove=True,
            )
            save_ex.assert_called_once()

            assert exchange.thread_id == proposal._thread_id
            assert exchange.initiator == V10PresentationExchange.INITIATOR_SELF
            assert exchange.role == V10PresentationExchange.ROLE_PROVER
            assert exchange.state == V10PresentationExchange.STATE_PROPOSAL_SENT
            assert exchange.auto_remove is True

    async def test_receive_proposal(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)
        proposal = PresentationProposal()

        with mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex:
            exchange = await self.manager.receive_proposal(proposal, connection_record)
            save_ex.assert_called_once()

            assert exchange.state == V10PresentationExchange.STATE_PROPOSAL_RECEIVED

    async def test_create_bound_request(self):
        comment = "comment"

        proposal = PresentationProposal(presentation_proposal=PRES_PREVIEW)
        exchange = V10PresentationExchange(
            presentation_proposal_dict=proposal.serialize(),
            role=V10PresentationExchange.ROLE_VERIFIER,
        )
        exchange.save = mock.CoroutineMock()
        (ret_exchange, pres_req_msg) = await self.manager.create_bound_request(
            presentation_exchange_record=exchange,
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            comment=comment,
        )
        assert ret_exchange is exchange
        exchange.save.assert_called_once()

    async def test_create_exchange_for_request(self):
        indy_proof_req = await PRES_PREVIEW.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )
        pres_req = PresentationRequest(
            request_presentations_attach=[
                AttachDecorator.data_base64(
                    mapping=indy_proof_req,
                    ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
                )
            ]
        )

        with mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex:
            exchange = await self.manager.create_exchange_for_request(
                CONN_ID,
                pres_req,
                auto_remove=True,
            )
            save_ex.assert_called_once()

            assert exchange.thread_id == pres_req._thread_id
            assert exchange.initiator == V10PresentationExchange.INITIATOR_SELF
            assert exchange.role == V10PresentationExchange.ROLE_VERIFIER
            assert exchange.state == V10PresentationExchange.STATE_REQUEST_SENT
            assert exchange.auto_remove is True

    async def test_receive_request(self):
        exchange_in = V10PresentationExchange()

        with mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex:
            exchange_out = await self.manager.receive_request(exchange_in)
            save_ex.assert_called_once()

            assert exchange_out.state == V10PresentationExchange.STATE_REQUEST_RECEIVED

    async def test_create_presentation(self):
        exchange_in = V10PresentationExchange()
        indy_proof_req = await PRES_PREVIEW.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )

        exchange_in.presentation_request = indy_proof_req

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_module, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                indy_proof_req, holder=self.holder
            )
            assert not req_creds["self_attested_attributes"]
            assert len(req_creds["requested_attributes"]) == 2
            assert len(req_creds["requested_predicates"]) == 1

            (exchange_out, pres_msg) = await self.manager.create_presentation(
                exchange_in, req_creds
            )
            save_ex.assert_called_once()
            assert exchange_out.state == V10PresentationExchange.STATE_PRESENTATION_SENT

    async def test_create_presentation_proof_req_non_revoc_interval_none(self):
        exchange_in = V10PresentationExchange()
        indy_proof_req = await PRES_PREVIEW.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )
        indy_proof_req["non_revoked"] = None  # simulate interop with indy-vcx

        exchange_in.presentation_request = indy_proof_req

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_module, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                indy_proof_req, holder=self.holder
            )
            assert not req_creds["self_attested_attributes"]
            assert len(req_creds["requested_attributes"]) == 2
            assert len(req_creds["requested_predicates"]) == 1

            (exchange_out, pres_msg) = await self.manager.create_presentation(
                exchange_in, req_creds
            )
            save_ex.assert_called_once()
            assert exchange_out.state == V10PresentationExchange.STATE_PRESENTATION_SENT

    async def test_create_presentation_self_asserted(self):
        PRES_PREVIEW_SELFIE = IndyPresPreview(
            attributes=[
                IndyPresAttrSpec(name="player", value="Richie Knucklez"),
                IndyPresAttrSpec(
                    name="screenCapture",
                    mime_type="image/png",
                    value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                ),
            ],
            predicates=[
                IndyPresPredSpec(
                    name="highScore",
                    cred_def_id=None,
                    predicate=">=",
                    threshold=1000000,
                )
            ],
        )

        exchange_in = V10PresentationExchange()
        indy_proof_req = await PRES_PREVIEW_SELFIE.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )

        exchange_in.presentation_request = indy_proof_req

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_module, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                indy_proof_req, holder=self.holder
            )
            assert len(req_creds["self_attested_attributes"]) == 3
            assert not req_creds["requested_attributes"]
            assert not req_creds["requested_predicates"]

            (exchange_out, pres_msg) = await self.manager.create_presentation(
                exchange_in, req_creds
            )
            save_ex.assert_called_once()
            assert exchange_out.state == V10PresentationExchange.STATE_PRESENTATION_SENT

    async def test_create_presentation_no_revocation(self):
        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=mock.MagicMock())
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {"revocation": None}}
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)

        exchange_in = V10PresentationExchange()
        indy_proof_req = await PRES_PREVIEW.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )

        exchange_in.presentation_request = indy_proof_req

        self.holder = mock.MagicMock(IndyHolder, autospec=True)
        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {"referent": "dummy_reft"},
                    "attrs": {
                        "player": "Richie Knucklez",
                        "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        "highScore": "1234560",
                    },
                },  # leave this comma: return a tuple
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds
        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "schema_id": S_ID,
                    "cred_def_id": CD_ID,
                    "rev_reg_id": None,
                    "cred_rev_id": None,
                }
            )
        )
        self.holder.create_presentation = mock.CoroutineMock(return_value="{}")
        self.profile.context.injector.bind_instance(IndyHolder, self.holder)

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_module, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module.LOGGER, "info", mock.MagicMock()
            ) as mock_log_info,
        ):
            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                indy_proof_req, holder=self.holder
            )

            (exchange_out, pres_msg) = await self.manager.create_presentation(
                exchange_in, req_creds
            )
            save_ex.assert_called_once()
            assert exchange_out.state == V10PresentationExchange.STATE_PRESENTATION_SENT

            # exercise superfluous timestamp removal
            for pred_reft_spec in req_creds["requested_predicates"].values():
                pred_reft_spec["timestamp"] = 1234567890
            await self.manager.create_presentation(exchange_in, req_creds)
            mock_log_info.assert_called_once()

    async def test_create_presentation_bad_revoc_state(self):
        exchange_in = V10PresentationExchange()
        indy_proof_req = await PRES_PREVIEW.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )

        exchange_in.presentation_request = indy_proof_req

        self.holder = mock.MagicMock(IndyHolder, autospec=True)
        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {"referent": "dummy_reft"},
                    "attrs": {
                        "player": "Richie Knucklez",
                        "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        "highScore": "1234560",
                    },
                },  # leave this comma: return a tuple
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds

        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "schema_id": S_ID,
                    "cred_def_id": CD_ID,
                    "rev_reg_id": RR_ID,
                    "cred_rev_id": 1,
                }
            )
        )
        self.holder.create_presentation = mock.CoroutineMock(return_value="{}")
        self.holder.create_revocation_state = mock.CoroutineMock(
            side_effect=IndyHolderError("Problem", {"message": "Nope"})
        )
        self.profile.context.injector.bind_instance(IndyHolder, self.holder)

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True),
            mock.patch.object(
                test_module, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                indy_proof_req, holder=self.holder
            )

            with self.assertRaises(IndyHolderError):
                await self.manager.create_presentation(exchange_in, req_creds)

    async def test_create_presentation_multi_matching_proposal_creds_names(self):
        exchange_in = V10PresentationExchange()
        indy_proof_req = await PRES_PREVIEW_NAMES.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )

        exchange_in.presentation_request = indy_proof_req

        self.holder = mock.MagicMock(IndyHolder, autospec=True)
        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {
                        "referent": "dummy_reft_0",
                        "cred_def_id": CD_ID,
                        "attrs": {
                            "player": "Richie Knucklez",
                            "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "highScore": "1234560",
                        },
                    }
                },
                {
                    "cred_info": {
                        "referent": "dummy_reft_1",
                        "cred_def_id": CD_ID,
                        "attrs": {
                            "player": "Richie Knucklez",
                            "screenCapture": "aW1hZ2luZSBhbm90aGVyIHNjcmVlbiBjYXB0dXJl",
                            "highScore": "1515880",
                        },
                    }
                },
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds
        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "schema_id": S_ID,
                    "cred_def_id": CD_ID,
                    "rev_reg_id": RR_ID,
                    "cred_rev_id": 1,
                }
            )
        )
        self.holder.create_presentation = mock.CoroutineMock(return_value="{}")
        self.holder.create_revocation_state = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "witness": {"omega": "1 ..."},
                    "rev_reg": {"accum": "21 ..."},
                    "timestamp": NOW,
                }
            )
        )
        self.profile.context.injector.bind_instance(IndyHolder, self.holder)

        more_magic_rr = mock.MagicMock(
            get_or_fetch_local_tails_path=mock.CoroutineMock(
                return_value="/tmp/sample/tails/path"
            )
        )
        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                test_module, "AttachDecorator", autospec=True
            ) as mock_attach_decorator,
            mock.patch.object(
                test_indy_util_module, "RevocationRegistry", autospec=True
            ) as mock_rr,
        ):
            mock_rr.from_definition = mock.MagicMock(return_value=more_magic_rr)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            req_creds = await indy_proof_req_preview2indy_requested_creds(
                indy_proof_req, preview=PRES_PREVIEW_NAMES, holder=self.holder
            )
            assert not req_creds["self_attested_attributes"]
            assert len(req_creds["requested_attributes"]) == 1
            assert len(req_creds["requested_predicates"]) == 1

            (exchange_out, pres_msg) = await self.manager.create_presentation(
                exchange_in, req_creds
            )
            save_ex.assert_called_once()
            assert exchange_out.state == V10PresentationExchange.STATE_PRESENTATION_SENT

    async def test_no_matching_creds_for_proof_req(self):
        indy_proof_req = await PRES_PREVIEW.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )
        get_creds = mock.CoroutineMock(return_value=())
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds

        with self.assertRaises(ValueError):
            await indy_proof_req_preview2indy_requested_creds(
                indy_proof_req, holder=self.holder
            )

        get_creds = mock.CoroutineMock(
            return_value=(
                {
                    "cred_info": {"referent": "dummy_reft"},
                    "attrs": {
                        "player": "Richie Knucklez",
                        "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        "highScore": "1234560",
                    },
                },  # leave this comma: return a tuple
            )
        )
        self.holder.get_credentials_for_presentation_request_by_referent = get_creds

    async def test_receive_presentation(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)

        exchange_dummy = V10PresentationExchange(
            presentation_proposal_dict={
                "presentation_proposal": {
                    "@type": DIDCommPrefix.qualify_current(
                        "present-proof/1.0/presentation-preview"
                    ),
                    "attributes": [
                        {
                            "name": "player",
                            "cred_def_id": CD_ID,
                            "value": "Richie Knucklez",
                        },
                        {
                            "name": "screenCapture",
                            "cred_def_id": CD_ID,
                            "value": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        },
                    ],
                    "predicates": [
                        {
                            "name": "highScore",
                            "cred_def_id": CD_ID,
                            "predicate": ">=",
                            "threshold": 1000000,
                        }
                    ],
                }
            },
            presentation_request={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {
                    "0_player_uuid": {
                        "name": "player",
                        "restrictions": [{"cred_def_id": CD_ID}],
                    },
                    "0_screencapture_uuid": {
                        "name": "screenCapture",
                        "restrictions": [{"cred_def_id": CD_ID}],
                    },
                },
                "requested_predicates": {
                    "0_highscore_GE_uuid": {
                        "name": "highScore",
                        "p_type": ">=",
                        "p_value": 1000000,
                        "restrictions": [{"cred_def_id": CD_ID}],
                    }
                },
            },
            presentation={
                "proof": {"proofs": []},
                "requested_proof": {
                    "revealed_attrs": {
                        "0_favourite_uuid": {
                            "sub_proof_index": 0,
                            "raw": "potato",
                            "encoded": "12345678901234567890",
                        },
                        "1_icon_uuid": {
                            "sub_proof_index": 1,
                            "raw": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "encoded": "12345678901234567890",
                        },
                    },
                    "self_attested_attrs": {},
                    "unrevealed_attrs": {},
                    "predicates": {},
                },
                "identifiers": [
                    {
                        "schema_id": S_ID,
                        "cred_def_id": CD_ID,
                        "rev_reg_id": None,
                        "timestamp": None,
                    },
                    {
                        "schema_id": S_ID,
                        "cred_def_id": CD_ID,
                        "rev_reg_id": None,
                        "timestamp": None,
                    },
                ],
            },
        )

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True),
            mock.patch.object(
                V10PresentationExchange,
                "retrieve_by_tag_filter",
                mock.CoroutineMock(return_value=exchange_dummy),
            ) as retrieve_ex,
        ):
            retrieve_ex.side_effect = [exchange_dummy]
            exchange_out = await self.manager.receive_presentation(
                PRES, connection_record, None
            )
            assert exchange_out.state == (
                V10PresentationExchange.STATE_PRESENTATION_RECEIVED
            )

    async def test_receive_presentation_oob(self):
        exchange_dummy = V10PresentationExchange(
            presentation_proposal_dict={
                "presentation_proposal": {
                    "@type": DIDCommPrefix.qualify_current(
                        "present-proof/1.0/presentation-preview"
                    ),
                    "attributes": [
                        {
                            "name": "player",
                            "cred_def_id": CD_ID,
                            "value": "Richie Knucklez",
                        },
                        {
                            "name": "screenCapture",
                            "cred_def_id": CD_ID,
                            "value": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                        },
                    ],
                    "predicates": [
                        {
                            "name": "highScore",
                            "cred_def_id": CD_ID,
                            "predicate": ">=",
                            "threshold": 1000000,
                        }
                    ],
                }
            },
            presentation_request={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {
                    "0_player_uuid": {
                        "name": "player",
                        "restrictions": [{"cred_def_id": CD_ID}],
                    },
                    "0_screencapture_uuid": {
                        "name": "screenCapture",
                        "restrictions": [{"cred_def_id": CD_ID}],
                    },
                },
                "requested_predicates": {
                    "0_highscore_GE_uuid": {
                        "name": "highScore",
                        "p_type": ">=",
                        "p_value": 1000000,
                        "restrictions": [{"cred_def_id": CD_ID}],
                    }
                },
            },
            presentation={
                "proof": {"proofs": []},
                "requested_proof": {
                    "revealed_attrs": {
                        "0_favourite_uuid": {
                            "sub_proof_index": 0,
                            "raw": "potato",
                            "encoded": "12345678901234567890",
                        },
                        "1_icon_uuid": {
                            "sub_proof_index": 1,
                            "raw": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "encoded": "12345678901234567890",
                        },
                    },
                    "self_attested_attrs": {},
                    "unrevealed_attrs": {},
                    "predicates": {},
                },
                "identifiers": [
                    {
                        "schema_id": S_ID,
                        "cred_def_id": CD_ID,
                        "rev_reg_id": None,
                        "timestamp": None,
                    },
                    {
                        "schema_id": S_ID,
                        "cred_def_id": CD_ID,
                        "rev_reg_id": None,
                        "timestamp": None,
                    },
                ],
            },
        )

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True),
            mock.patch.object(
                V10PresentationExchange, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.side_effect = [exchange_dummy]
            exchange_out = await self.manager.receive_presentation(PRES, None, None)
            assert exchange_out.state == (
                V10PresentationExchange.STATE_PRESENTATION_RECEIVED
            )

    async def test_receive_presentation_bait_and_switch(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)

        exchange_dummy = V10PresentationExchange(
            presentation_proposal_dict={
                "presentation_proposal": {
                    "@type": DIDCommPrefix.qualify_current(
                        "present-proof/1.0/presentation-preview"
                    ),
                    "attributes": [
                        {
                            "name": "player",
                            "cred_def_id": CD_ID,
                            "value": "Richie Knucklez",
                        },
                        {
                            "name": "screenCapture",
                            "cred_def_id": CD_ID,
                            "value": "YSBwaWN0dXJlIG9mIGEgcG90YXRv",
                        },
                    ],
                    "predicates": [
                        {
                            "name": "highScore",
                            "cred_def_id": CD_ID,
                            "predicate": ">=",
                            "threshold": 1000000,
                        }
                    ],
                }
            },
            presentation_request={
                "name": "proof-request",
                "version": "1.0",
                "nonce": "1234567890",
                "requested_attributes": {
                    "0_player_uuid": {
                        "name": "player",
                        "restrictions": [{"cred_def_id": CD_ID}],
                    },
                    "0_screencapture_uuid": {
                        "name": "screenCapture",
                        "restrictions": [{"cred_def_id": CD_ID}],
                    },
                },
                "requested_predicates": {
                    "0_highscore_GE_uuid": {
                        "name": "highScore",
                        "p_type": ">=",
                        "p_value": 1000000,
                        "restrictions": [{"cred_def_id": CD_ID}],
                    }
                },
            },
            presentation={
                "proof": {"proofs": []},
                "requested_proof": {
                    "revealed_attrs": {
                        "0_favourite_uuid": {
                            "sub_proof_index": 0,
                            "raw": "potato",
                            "encoded": "12345678901234567890",
                        },
                        "1_icon_uuid": {
                            "sub_proof_index": 1,
                            "raw": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "encoded": "12345678901234567890",
                        },
                    },
                    "self_attested_attrs": {},
                    "unrevealed_attrs": {},
                    "predicates": {},
                },
                "identifiers": [
                    {
                        "schema_id": S_ID,
                        "cred_def_id": CD_ID,
                        "rev_reg_id": None,
                        "timestamp": None,
                    },
                    {
                        "schema_id": S_ID,
                        "cred_def_id": CD_ID,
                        "rev_reg_id": None,
                        "timestamp": None,
                    },
                ],
            },
        )

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True),
            mock.patch.object(
                V10PresentationExchange, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = exchange_dummy
            with self.assertRaises(PresentationManagerError):
                await self.manager.receive_presentation(PRES, connection_record, None)

    async def test_receive_presentation_connectionless(self):
        exchange_dummy = V10PresentationExchange()

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                V10PresentationExchange, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = exchange_dummy
            exchange_out = await self.manager.receive_presentation(PRES, None, None)
            save_ex.assert_called_once()

            assert exchange_out.state == (
                V10PresentationExchange.STATE_PRESENTATION_RECEIVED
            )

    async def test_verify_presentation(self):
        indy_proof_req = await PRES_PREVIEW.indy_proof_request(
            name=PROOF_REQ_NAME,
            version=PROOF_REQ_VERSION,
            nonce=PROOF_REQ_NONCE,
            profile=self.profile,
        )
        pres_req = PresentationRequest(
            request_presentations_attach=[
                AttachDecorator.data_base64(
                    mapping=indy_proof_req,
                    ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
                )
            ]
        )
        exchange_in = V10PresentationExchange(
            presentation_exchange_id="dummy-pxid",
            connection_id="dummy-conn-id",
            initiator=V10PresentationExchange.INITIATOR_SELF,
            role=V10PresentationExchange.ROLE_VERIFIER,
            presentation_request=pres_req,
            presentation=INDY_PROOF,
        )

        with mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex:
            exchange_out = await self.manager.verify_presentation(exchange_in)
            save_ex.assert_called_once()

            assert exchange_out.state == (V10PresentationExchange.STATE_VERIFIED)

    async def test_send_presentation_ack(self):
        exchange = V10PresentationExchange(connection_id="dummy")

        responder = MockResponder()
        self.profile.context.injector.bind_instance(BaseResponder, responder)

        await self.manager.send_presentation_ack(exchange)
        messages = responder.messages
        assert len(messages) == 1

    async def test_send_presentation_ack_oob(self):
        exchange = V10PresentationExchange(thread_id="some-thread-id")

        responder = MockResponder()
        self.profile.context.injector.bind_instance(BaseResponder, responder)

        with mock.patch.object(
            test_module.OobRecord, "retrieve_by_tag_filter"
        ) as mock_retrieve_oob:
            await self.manager.send_presentation_ack(exchange)
            messages = responder.messages
            assert len(messages) == 1
            assert mock_retrieve_oob.called

    async def test_send_presentation_ack_no_responder(self):
        exchange = V10PresentationExchange()

        self.profile.context.injector.clear_binding(BaseResponder)

        with mock.patch.object(
            test_module.OobRecord, "retrieve_by_tag_filter"
        ) as mock_retrieve_oob:
            await self.manager.send_presentation_ack(exchange)
            assert mock_retrieve_oob.called

    async def test_receive_presentation_ack_a(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)

        exchange_dummy = V10PresentationExchange()
        message = mock.MagicMock()

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                V10PresentationExchange, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = exchange_dummy
            exchange_out = await self.manager.receive_presentation_ack(
                message, connection_record
            )
            save_ex.assert_called_once()

            assert exchange_out.state == (
                V10PresentationExchange.STATE_PRESENTATION_ACKED
            )

    async def test_receive_presentation_ack_b(self):
        connection_record = mock.MagicMock(connection_id=CONN_ID)

        exchange_dummy = V10PresentationExchange()
        message = mock.MagicMock(_verification_result="true")

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                V10PresentationExchange, "retrieve_by_tag_filter", autospec=True
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = exchange_dummy
            exchange_out = await self.manager.receive_presentation_ack(
                message, connection_record
            )
            save_ex.assert_called_once()

            assert exchange_out.state == (
                V10PresentationExchange.STATE_PRESENTATION_ACKED
            )
            assert exchange_out.verified == "true"

    async def test_receive_problem_report(self):
        connection_id = "connection-id"
        stored_exchange = V10PresentationExchange(
            presentation_exchange_id="dummy-pxid",
            connection_id=connection_id,
            initiator=V10PresentationExchange.INITIATOR_SELF,
            role=V10PresentationExchange.ROLE_VERIFIER,
            state=V10PresentationExchange.STATE_PROPOSAL_RECEIVED,
            thread_id="dummy-thid",
        )
        problem = PresentationProblemReport(
            description={
                "code": test_module.ProblemReportReason.ABANDONED.value,
                "en": "Change of plans",
            }
        )

        with (
            mock.patch.object(V10PresentationExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                V10PresentationExchange,
                "retrieve_by_tag_filter",
                mock.CoroutineMock(),
            ) as retrieve_ex,
            mock.patch.object(
                self.profile,
                "session",
                mock.MagicMock(return_value=self.profile.session()),
            ) as session,
        ):
            retrieve_ex.return_value = stored_exchange

            ret_exchange = await self.manager.receive_problem_report(
                problem, connection_id
            )
            retrieve_ex.assert_called_once_with(
                session.return_value,
                {"thread_id": problem._thread_id},
                {"connection_id": connection_id},
            )
            save_ex.assert_called_once()

            assert ret_exchange.state == V10CredentialExchange.STATE_ABANDONED

    async def test_receive_problem_report_x(self):
        connection_id = "connection-id"
        stored_exchange = V10PresentationExchange(
            presentation_exchange_id="dummy-pxid",
            connection_id=connection_id,
            initiator=V10PresentationExchange.INITIATOR_SELF,
            role=V10PresentationExchange.ROLE_VERIFIER,
            state=V10PresentationExchange.STATE_PROPOSAL_RECEIVED,
            thread_id="dummy-thid",
        )
        problem = PresentationProblemReport(
            description={
                "code": test_module.ProblemReportReason.ABANDONED.value,
                "en": "Change of plans",
            }
        )

        with mock.patch.object(
            V10PresentationExchange,
            "retrieve_by_tag_filter",
            mock.CoroutineMock(),
        ) as retrieve_ex:
            retrieve_ex.side_effect = test_module.StorageNotFoundError("No such record")

            with self.assertRaises(test_module.StorageNotFoundError):
                await self.manager.receive_problem_report(problem, connection_id)



"""KeyInfo, DIDInfo."""

from typing import NamedTuple, Union, List, Optional

from .did_method import DIDMethod
from .key_type import KeyType

INVITATION_REUSE_KEY = "invitation_reuse"


class KeyInfo(NamedTuple):
    """Class returning key information."""

    verkey: str
    metadata: dict
    key_type: KeyType
    kid: Optional[Union[List[str], str]] = None


DIDInfo = NamedTuple(
    "DIDInfo",
    [
        ("did", str),
        ("verkey", str),
        ("metadata", dict),
        ("method", DIDMethod),
        ("key_type", KeyType),
    ],
)



"""Driver-based action menu service classes."""

import logging
from typing import Optional

from ....connections.models.conn_record import ConnRecord
from ....core.profile import Profile
from ....messaging.agent_message import AgentMessage
from .base_service import BaseMenuService
from .messages.menu import Menu

LOGGER = logging.getLogger(__name__)


class DriverMenuService(BaseMenuService):
    """Driver-based action menu service."""

    async def get_active_menu(
        self,
        profile: Profile,
        connection: Optional[ConnRecord] = None,
        thread_id: Optional[str] = None,
    ) -> Menu:
        """Render the current menu.

        Args:
            profile: The profile
            connection: The active connection record
            thread_id: The thread identifier from the requesting message.
        """
        await profile.notify(
            "acapy::actionmenu::get-active-menu",
            {
                "connection_id": connection and connection.connection_id,
                "thread_id": thread_id,
            },
        )
        return None

    async def perform_menu_action(
        self,
        profile: Profile,
        action_name: str,
        action_params: dict,
        connection: Optional[ConnRecord] = None,
        thread_id: Optional[str] = None,
    ) -> AgentMessage:
        """Perform an action defined by the active menu.

        Args:
            profile: The profile
            action_name: The unique name of the action being performed
            action_params: A collection of parameters for the action
            connection: The active connection record
            thread_id: The thread identifier from the requesting message.
        """
        await profile.notify(
            "acapy::actionmenu::perform-menu-action",
            {
                "connection_id": connection and connection.connection_id,
                "thread_id": thread_id,
                "action_name": action_name,
                "action_params": action_params,
            },
        )
        return None



from unittest import IsolatedAsyncioTestCase

from ......messaging.base_handler import HandlerException
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......utils.testing import create_test_profile
from .....out_of_band.v1_0.messages.invitation import InvitationMessage, Service
from ...messages.invitation import Invitation
from .. import invitation_handler as test_module

TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
TEST_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_ROUTE_VERKEY = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"
TEST_LABEL = "Label"
TEST_ENDPOINT = "http://localhost"
TEST_IMAGE_URL = "http://aries.ca/images/sample.png"


class TestInvitationHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.context = RequestContext.test_context(await create_test_profile())
        self.context.connection_ready = True
        service = Service(
            did=TEST_DID,
            recipient_keys=[TEST_VERKEY],
            service_endpoint=TEST_ENDPOINT,
            routing_keys=[TEST_ROUTE_VERKEY],
        )
        self.context.message = Invitation(
            invitation=InvitationMessage(
                label=TEST_LABEL,
                image_url=TEST_IMAGE_URL,
                services=[service],
            ),
            message="Hello World",
        )
        self.context.connection_record = mock.MagicMock(connection_id="dummy")

    async def test_handle(self):
        handler = test_module.InvitationHandler()

        responder = MockResponder()
        with mock.patch.object(
            self.context, "inject_or", mock.MagicMock()
        ) as mock_ctx_inject:
            mock_ctx_inject.return_value = mock.MagicMock(
                return_invitation=mock.CoroutineMock()
            )

            await handler.handle(self.context, responder)

            mock_ctx_inject.return_value.return_invitation.assert_called_once_with(
                self.context.connection_record.connection_id,
                self.context.message,
                mock.ANY,
                responder.send,
            )

    async def test_handle_no_service(self):
        handler = test_module.InvitationHandler()

        with self.assertRaises(HandlerException):
            await handler.handle(self.context, None)

    async def test_handle_not_ready(self):
        handler = test_module.InvitationHandler()
        self.context.connection_ready = False

        with self.assertRaises(HandlerException):
            await handler.handle(self.context, None)



"""Test RevNotificationRecord."""

import pytest
import pytest_asyncio

from ......storage.error import StorageDuplicateError, StorageNotFoundError
from ......utils.testing import create_test_profile
from ...messages.revoke import Revoke
from ..rev_notification_record import RevNotificationRecord


@pytest_asyncio.fixture
async def profile():
    profile = await create_test_profile()
    yield profile


@pytest.fixture
def rec():
    yield RevNotificationRecord(
        rev_reg_id="mock_rev_reg_id",
        cred_rev_id="mock_cred_rev_id",
        connection_id="mock_connection_id",
        thread_id="mock_thread_id",
        comment="mock_comment",
        version="v2_0",
    )


@pytest.mark.asyncio
async def test_storage(profile, rec):
    async with profile.session() as session:
        await rec.save(session)
        recalled = await RevNotificationRecord.retrieve_by_id(
            session, rec.revocation_notification_id
        )
        assert recalled == rec
        recalled = await RevNotificationRecord.query_by_ids(
            session, cred_rev_id="mock_cred_rev_id", rev_reg_id="mock_rev_reg_id"
        )
        assert recalled == rec
        [recalled] = await RevNotificationRecord.query_by_rev_reg_id(
            session, rev_reg_id="mock_rev_reg_id"
        )
        assert recalled == rec

        with pytest.raises(StorageNotFoundError):
            await RevNotificationRecord.query_by_ids(
                session, cred_rev_id="unknown", rev_reg_id="unknown"
            )

        with pytest.raises(StorageDuplicateError):
            another = RevNotificationRecord(
                rev_reg_id="mock_rev_reg_id",
                cred_rev_id="mock_cred_rev_id",
                version="v2_0",
            )
            await another.save(session)
            await RevNotificationRecord.query_by_ids(
                session, cred_rev_id="mock_cred_rev_id", rev_reg_id="mock_rev_reg_id"
            )


def test_to_message(rec):
    message = rec.to_message()
    assert isinstance(message, Revoke)
    assert message.credential_id == f"{rec.rev_reg_id}::{rec.cred_rev_id}"
    assert message.comment == rec.comment

    with pytest.raises(ValueError):
        rec.thread_id = None
        rec.to_message()



from datetime import datetime, timedelta
from unittest import IsolatedAsyncioTestCase, mock

from .....messaging.util import datetime_to_str
from ..proof_purpose import ProofPurpose


class TestProofPurpose(IsolatedAsyncioTestCase):
    async def test_properties(self):
        term = "TestTerm"
        date = datetime.now()
        delta = timedelta(1)
        proof_purpose = ProofPurpose(term=term, date=date, max_timestamp_delta=delta)
        proof_purpose2 = ProofPurpose(term=term, date=date, max_timestamp_delta=delta)

        assert proof_purpose.term == term
        assert proof_purpose.date == date
        assert proof_purpose.max_timestamp_delta == delta

        assert proof_purpose2 == proof_purpose
        assert proof_purpose != 10

    async def test_validate(self):
        proof_purpose = ProofPurpose(term="ProofTerm", date=datetime.now())

        result = proof_purpose.validate(
            proof=mock.MagicMock(),
            document=mock.MagicMock(),
            suite=mock.MagicMock(),
            verification_method=mock.MagicMock(),
            document_loader=mock.MagicMock(),
        )
        assert result.valid

    async def test_validate_timestamp_delta(self):
        date = datetime.now()
        proof_purpose = ProofPurpose(
            term="ProofTerm", date=date, max_timestamp_delta=timedelta(10)
        )

        result = proof_purpose.validate(
            proof={"created": datetime_to_str(date + timedelta(5))},
            document=mock.MagicMock(),
            suite=mock.MagicMock(),
            verification_method=mock.MagicMock(),
            document_loader=mock.MagicMock(),
        )
        assert result.valid

    async def test_validate_timestamp_delta_out_of_rage(self):
        date = datetime.now()
        proof_purpose = ProofPurpose(
            term="ProofTerm", date=date, max_timestamp_delta=timedelta(10)
        )

        result = proof_purpose.validate(
            proof={"created": datetime_to_str(date + timedelta(15))},
            document=mock.MagicMock(),
            suite=mock.MagicMock(),
            verification_method=mock.MagicMock(),
            document_loader=mock.MagicMock(),
        )

        assert not result.valid



"""Represents an invitation returned to the introduction service."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ....out_of_band.v1_0.messages.invitation import (
    InvitationMessage,
    InvitationMessageSchema,
)
from ..message_types import INVITATION, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.invitation_handler.InvitationHandler"


class Invitation(AgentMessage):
    """Class representing an invitation returned to the introduction service."""

    class Meta:
        """Metadata for an invitation."""

        handler_class = HANDLER_CLASS
        message_type = INVITATION
        schema_class = "InvitationSchema"

    def __init__(
        self,
        *,
        invitation: Optional[InvitationMessage] = None,
        message: Optional[str] = None,
        **kwargs,
    ):
        """Initialize invitation object.

        Args:
            invitation: The connection invitation
            message: Comments on the introduction
            kwargs: Additional key word arguments for the message
        """
        super().__init__(**kwargs)
        self.invitation = invitation
        self.message = message


class InvitationSchema(AgentMessageSchema):
    """Invitation request schema class."""

    class Meta:
        """Invitation request schema metadata."""

        model_class = Invitation
        unknown = EXCLUDE

    invitation = fields.Nested(InvitationMessageSchema(), required=True)
    message = fields.Str(
        required=False,
        allow_none=True,
        metadata={
            "description": "Comments on the introduction",
            "example": "Hello Bob, it's Charlie as Alice mentioned",
        },
    )



"""Base Class for DID Resolvers."""

import re
import warnings
from abc import ABC, abstractmethod
from enum import Enum
from typing import NamedTuple, Optional, Pattern, Sequence, Text, Union

from pydid import DID

from ..cache.base import BaseCache
from ..config.injection_context import InjectionContext
from ..core.error import BaseError
from ..core.profile import Profile


class ResolverError(BaseError):
    """Base class for resolver exceptions."""


class DIDNotFound(ResolverError):
    """Raised when DID is not found in verifiable data registry."""


class DIDMethodNotSupported(ResolverError):
    """Raised when no resolver is registered for a given did method."""


class ResolverType(Enum):
    """Resolver Type declarations."""

    NATIVE = "native"
    NON_NATIVE = "non-native"


class ResolutionMetadata(NamedTuple):
    """Resolution Metadata."""

    resolver_type: ResolverType
    resolver: str
    retrieved_time: str
    duration: int

    def serialize(self) -> dict:
        """Return serialized resolution metadata."""
        return {**self._asdict(), "resolver_type": self.resolver_type.value}


class ResolutionResult:
    """Resolution Class to pack the DID Doc and the resolution information."""

    def __init__(self, did_document: dict, metadata: ResolutionMetadata):
        """Initialize Resolution.

        Args:
            did_document: DID Document resolved
            metadata: Resolving details
        """
        self.did_document = did_document
        self.metadata = metadata

    def serialize(self) -> dict:
        """Return serialized resolution result."""
        return {
            "did_document": self.did_document,
            "metadata": self.metadata.serialize(),
        }


class BaseDIDResolver(ABC):
    """Base Class for DID Resolvers."""

    DEFAULT_TTL = 3600

    def __init__(self, type_: Optional[ResolverType] = None):
        """Initialize BaseDIDResolver.

        Args:
            type_ (Type): Type of resolver, native or non-native
        """
        self.type = type_ or ResolverType.NON_NATIVE

    @abstractmethod
    async def setup(self, context: InjectionContext):
        """Do asynchronous resolver setup."""

    @property
    def native(self):
        """Return if this resolver is native."""
        return self.type == ResolverType.NATIVE

    @property
    def supported_methods(self) -> Sequence[str]:
        """Return supported methods.

        DEPRECATED: Use supported_did_regex instead.
        """
        return []

    @property
    def supported_did_regex(self) -> Pattern:
        """Supported DID regex for matching this resolver to DIDs it can resolve.

        Override this property with a class var or similar to use regex
        matching on DIDs to determine if this resolver supports a given DID.
        """
        raise NotImplementedError(
            "supported_did_regex must be overridden by subclasses of BaseResolver "
            "to use default supports method"
        )

    async def supports(self, profile: Profile, did: str) -> bool:
        """Return if this resolver supports the given DID.

        Override this method to determine if this resolver supports a DID based
        on information other than just a regular expression; i.e. check a value
        in storage, query a resolver connection record, etc.
        """
        try:
            supported_did_regex = self.supported_did_regex
        except NotImplementedError as error:
            methods = self.supported_methods
            if not methods:
                raise error
            warnings.warn(
                "BaseResolver.supported_methods is deprecated; "
                "use supported_did_regex instead",
                DeprecationWarning,
            )

            supported_did_regex = re.compile("^did:(?:{}):.*$".format("|".join(methods)))

        return bool(supported_did_regex.match(did))

    async def resolve(
        self,
        profile: Profile,
        did: Union[str, DID],
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve a DID using this resolver.

        Handles caching of results.
        """
        if isinstance(did, DID):
            did = str(did)
        else:
            DID.validate(did)
        if not await self.supports(profile, did):
            raise DIDMethodNotSupported(
                f"{self.__class__.__name__} does not support DID method for: {did}"
            )

        cache_key = f"resolver::{type(self).__name__}::{did}"
        cache = profile.inject_or(BaseCache)
        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    return entry.result
                else:
                    result = await self._resolve(profile, did, service_accept)
                    await entry.set_result(result, ttl=self.DEFAULT_TTL)
                    return result

        return await self._resolve(profile, did, service_accept)

    @abstractmethod
    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve a DID using this resolver."""



"""Model for representing a stored verifiable credential."""

import logging
from typing import Mapping, Optional, Sequence

from marshmallow import EXCLUDE, fields
from uuid_utils import uuid4

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import ENDPOINT_EXAMPLE, ENDPOINT_VALIDATE, UUID4_EXAMPLE

LOGGER = logging.getLogger(__name__)


class VCRecord(BaseModel):
    """Verifiable credential storage record class."""

    class Meta:
        """VCRecord metadata."""

        schema_class = "VCRecordSchema"

    def __init__(
        self,
        *,
        contexts: Sequence[str],  # context is required by spec
        expanded_types: Sequence[str],  # expanded type from contexts and types
        issuer_id: str,  # issuer ID is required by spec
        subject_ids: Sequence[str],  # one or more subject IDs may be present
        schema_ids: Sequence[str],  # one or more credential schema IDs may be present
        proof_types: Sequence[str],  # one or more proof types may be present
        cred_value: Mapping,  # the credential value as a JSON-serializable mapping
        given_id: Optional[str] = None,  # value of the credential 'id' property, if any
        cred_tags: Optional[
            Mapping
        ] = None,  # tags for retrieval (derived from attribute values)
        record_id: Optional[str] = None,  # specify the storage record ID
    ):
        """Initialize some defaults on record."""
        super().__init__()
        self.contexts = set(contexts) if contexts else set()
        self.expanded_types = set(expanded_types) if expanded_types else set()
        self.schema_ids = set(schema_ids) if schema_ids else set()
        self.issuer_id = issuer_id
        self.subject_ids = set(subject_ids) if subject_ids else set()
        self.proof_types = set(proof_types) if proof_types else set()
        self.cred_value = cred_value
        self.given_id = given_id
        self.cred_tags = cred_tags or {}
        self.record_id = record_id or uuid4().hex

    def serialize(self, as_string=False) -> dict:
        """Create a JSON-compatible dict representation of the model instance.

        Args:
            as_string: Return a string of JSON instead of a dict

        Returns:
            A dict representation of this model, or a JSON string if as_string is True

        """

        list_coercion = VCRecord(**dict(vars(self).items()))
        for k, v in vars(self).items():
            if isinstance(v, set):
                setattr(list_coercion, k, list(v))

        return super(VCRecord, list_coercion).serialize(as_string=as_string)

    def __eq__(self, other: object) -> bool:
        """Compare two VC records for equality."""
        if not isinstance(other, VCRecord):
            return False
        return (
            other.contexts == self.contexts
            and other.expanded_types == self.expanded_types
            and other.subject_ids == self.subject_ids
            and other.schema_ids == self.schema_ids
            and other.issuer_id == self.issuer_id
            and other.proof_types == self.proof_types
            and other.given_id == self.given_id
            and other.record_id == self.record_id
            and other.cred_tags == self.cred_tags
            and other.cred_value == self.cred_value
        )


class VCRecordSchema(BaseModelSchema):
    """Verifiable credential storage record schema class."""

    class Meta:
        """Verifiable credential storage record schema metadata."""

        model_class = VCRecord
        unknown = EXCLUDE

    contexts = fields.List(
        fields.Str(
            validate=ENDPOINT_VALIDATE,
            metadata={"description": "Context", "example": ENDPOINT_EXAMPLE},
        )
    )
    expanded_types = fields.List(
        fields.Str(
            metadata={
                "description": "JSON-LD expanded type extracted from type and context",
                "example": "https://w3id.org/citizenship#PermanentResidentCard",
            }
        )
    )
    schema_ids = fields.List(
        fields.Str(
            metadata={
                "description": "Schema identifier",
                "example": "https://example.org/examples/degree.json",
            }
        )
    )
    issuer_id = fields.Str(
        metadata={
            "description": "Issuer identifier",
            "example": "https://example.edu/issuers/14",
        }
    )
    subject_ids = fields.List(
        fields.Str(
            metadata={
                "description": "Subject identifier",
                "example": "did:example:ebfeb1f712ebc6f1c276e12ec21",
            }
        )
    )
    proof_types = fields.List(
        fields.Str(
            metadata={
                "description": "Signature suite used for proof",
                "example": "Ed25519Signature2018",
            }
        )
    )

    cred_value = fields.Dict(
        metadata={"description": "(JSON-serializable) credential value"}
    )
    given_id = fields.Str(
        metadata={
            "description": "Credential identifier",
            "example": "http://example.edu/credentials/3732",
        }
    )
    cred_tags = fields.Dict(
        keys=fields.Str(metadata={"description": "Retrieval tag name"}),
        values=fields.Str(metadata={"description": "Retrieval tag value"}),
    )
    record_id = fields.Str(
        metadata={"description": "Record identifier", "example": UUID4_EXAMPLE}
    )



import json
from unittest import IsolatedAsyncioTestCase

from ....connections.models.connection_target import ConnectionTarget
from ....tests import mock
from ....utils.testing import create_test_profile
from ...wire_format import BaseWireFormat
from .. import manager as test_module
from ..manager import (
    OutboundDeliveryError,
    OutboundTransportManager,
    OutboundTransportRegistrationError,
    QueuedOutboundMessage,
)
from ..message import OutboundMessage


class TestOutboundTransportManager(IsolatedAsyncioTestCase):
    async def test_register_path(self):
        mgr = OutboundTransportManager(await create_test_profile())
        mgr.register("http")
        assert mgr.get_registered_transport_for_scheme("http")
        assert mgr.MAX_RETRY_COUNT == 4

        assert mgr.get_registered_transport_for_scheme("xmpp") is None

        with self.assertRaises(OutboundTransportRegistrationError):
            mgr.register("http")

        with self.assertRaises(OutboundTransportRegistrationError):
            mgr.register("no.such.module.path")

    async def test_maximum_retry_count(self):
        self.profile = await create_test_profile({"transport.max_outbound_retry": 5})
        mgr = OutboundTransportManager(self.profile)
        mgr.register("http")
        assert mgr.MAX_RETRY_COUNT == 5

    async def test_setup(self):
        self.profile = await create_test_profile({"transport.outbound_configs": ["http"]})
        mgr = OutboundTransportManager(self.profile)
        with mock.patch.object(mgr, "register") as mock_register:
            await mgr.setup()
            mock_register.assert_called_once_with("http")

    async def test_send_message(self):
        self.profile = await create_test_profile()
        mgr = OutboundTransportManager(self.profile)

        transport_cls = mock.Mock(spec=[])
        with self.assertRaises(OutboundTransportRegistrationError):
            mgr.register_class(transport_cls, "transport_cls")

        transport = mock.MagicMock()
        transport.handle_message = mock.CoroutineMock()
        transport.wire_format.encode_message = mock.CoroutineMock()
        transport.start = mock.CoroutineMock()
        transport.stop = mock.CoroutineMock()
        transport.schemes = ["http"]
        transport.is_external = False

        transport_cls = mock.MagicMock()
        transport_cls.schemes = ["http"]
        transport_cls.return_value = transport
        mgr.register_class(transport_cls, "transport_cls")
        assert mgr.get_registered_transport_for_scheme("http") == "transport_cls"

        await mgr.start()
        await mgr.task_queue
        transport.start.assert_awaited_once_with()
        assert mgr.get_running_transport_for_scheme("http") == "transport_cls"

        message = OutboundMessage(payload="{}")
        assert "payload" in str(message)
        message.target = ConnectionTarget(
            endpoint="http://localhost",
            recipient_keys=[1, 2],
            routing_keys=[3],
            sender_key=4,
        )

        self.profile = await create_test_profile()
        await mgr.enqueue_message(self.profile, message)
        await mgr.flush()

        transport.wire_format.encode_message.assert_awaited_once()
        transport.handle_message.assert_awaited_once()

        with self.assertRaises(OutboundDeliveryError):
            mgr.get_running_transport_for_endpoint("localhost")

        message.target = ConnectionTarget(
            endpoint="localhost",
            recipient_keys=[1, 2],
            routing_keys=[3],
            sender_key=4,
        )
        with self.assertRaises(OutboundDeliveryError) as context:
            await mgr.enqueue_message(self.profile, message)
        assert "No supported transport" in str(context.exception)

        await mgr.stop()

        assert mgr.get_running_transport_for_scheme("http") is None
        transport.stop.assert_awaited_once_with()

    async def test_stop_cancel(self):
        self.profile = await create_test_profile({"transport.outbound_configs": ["http"]})
        mgr = OutboundTransportManager(self.profile)
        mgr._process_task = mock.MagicMock(
            done=mock.MagicMock(return_value=False), cancel=mock.MagicMock()
        )
        mgr.running_transports = {}
        await mgr.stop()
        mgr._process_task.cancel.assert_called_once()

    async def test_enqueue_webhook(self):
        self.profile = await create_test_profile()
        mgr = OutboundTransportManager(self.profile)
        test_topic = "test-topic"
        test_payload = {"test": "payload"}
        test_endpoint_host = "http://example"
        test_endpoint = f"{test_endpoint_host}#abc123"
        test_attempts = 2

        with self.assertRaises(OutboundDeliveryError):
            mgr.enqueue_webhook(
                test_topic, test_payload, test_endpoint, max_attempts=test_attempts
            )

        transport_cls = mock.MagicMock()
        transport_cls.schemes = ["http"]
        transport_cls.return_value = mock.MagicMock()
        transport_cls.return_value.schemes = ["http"]
        transport_cls.return_value.start = mock.CoroutineMock()
        tid = mgr.register_class(transport_cls, "transport_cls")
        await mgr.start_transport(tid)

        with mock.patch.object(mgr, "process_queued") as mock_process:
            mgr.enqueue_webhook(
                test_topic, test_payload, test_endpoint, max_attempts=test_attempts
            )
            mock_process.assert_called_once_with()
            assert len(mgr.outbound_new) == 1
            queued = mgr.outbound_new[0]
            assert queued.endpoint == f"{test_endpoint_host}/topic/{test_topic}/"
            assert json.loads(queued.payload) == test_payload
            assert queued.retries == test_attempts - 1
            assert queued.state == QueuedOutboundMessage.STATE_PENDING

    async def test_process_done_x(self):
        mock_task = mock.MagicMock(
            done=mock.MagicMock(return_value=True),
            exception=mock.MagicMock(return_value=KeyError("No such key")),
        )
        self.profile = await create_test_profile()
        mgr = OutboundTransportManager(self.profile)

        with mock.patch.object(
            mgr, "_process_task", mock.MagicMock()
        ) as mock_mgr_process:
            mock_mgr_process.done = mock.MagicMock(return_value=True)
            mgr._process_done(mock_task)

    async def test_process_finished_x(self):
        mock_queued = mock.MagicMock(retries=1)
        mock_task = mock.MagicMock(
            exc_info=(KeyError, KeyError("nope"), None),
        )
        self.profile = await create_test_profile()
        mgr = OutboundTransportManager(self.profile)

        with mock.patch.object(mgr, "process_queued", mock.MagicMock()):
            mgr.finished_encode(mock_queued, mock_task)
            mgr.finished_deliver(mock_queued, mock_task)
            mgr.finished_deliver(mock_queued, mock_task)

    async def test_process_loop_retry_now(self):
        mock_queued = mock.MagicMock(
            state=QueuedOutboundMessage.STATE_RETRY,
            retry_at=test_module.get_timer() - 1,
        )

        self.profile = await create_test_profile()
        mock_handle_not_delivered = mock.MagicMock()
        mgr = OutboundTransportManager(self.profile, mock_handle_not_delivered)
        mgr.outbound_buffer.append(mock_queued)

        with mock.patch.object(
            test_module, "trace_event", mock.MagicMock()
        ) as mock_trace:
            mock_trace.side_effect = KeyError()
            with self.assertRaises(KeyError):  # cover retry logic and bail
                await mgr._process_loop()
            assert mock_queued.retry_at is None

    async def test_process_loop_retry_later(self):
        mock_queued = mock.MagicMock(
            state=QueuedOutboundMessage.STATE_RETRY,
            retry_at=test_module.get_timer() + 3600,
        )

        self.profile = self.profile = await create_test_profile()
        mock_handle_not_delivered = mock.MagicMock()
        mgr = OutboundTransportManager(self.profile, mock_handle_not_delivered)
        mgr.outbound_buffer.append(mock_queued)

        with mock.patch.object(
            test_module.asyncio, "sleep", mock.CoroutineMock()
        ) as mock_sleep_x:
            mock_sleep_x.side_effect = KeyError()
            with self.assertRaises(KeyError):  # cover retry logic and bail
                await mgr._process_loop()
            assert mock_queued.retry_at is not None

    async def test_process_loop_new(self):
        self.profile = await create_test_profile()
        mock_handle_not_delivered = mock.MagicMock()
        mgr = OutboundTransportManager(self.profile, mock_handle_not_delivered)

        mgr.outbound_new = [
            mock.MagicMock(
                state=test_module.QueuedOutboundMessage.STATE_NEW,
                message=mock.MagicMock(enc_payload=b"encr"),
            )
        ]
        with (
            mock.patch.object(mgr, "deliver_queued_message", mock.MagicMock()),
            mock.patch.object(
                mgr.outbound_event, "wait", mock.CoroutineMock()
            ) as mock_wait,
            mock.patch.object(test_module, "trace_event", mock.MagicMock()),
        ):
            mock_wait.side_effect = KeyError()  # cover state=NEW logic and bail

            with self.assertRaises(KeyError):
                await mgr._process_loop()

    async def test_process_loop_new_deliver(self):
        self.profile = await create_test_profile()
        mock_handle_not_delivered = mock.MagicMock()
        mgr = OutboundTransportManager(self.profile, mock_handle_not_delivered)

        mgr.outbound_new = [
            mock.MagicMock(
                state=test_module.QueuedOutboundMessage.STATE_DELIVER,
                message=mock.MagicMock(enc_payload=b"encr"),
            )
        ]
        with (
            mock.patch.object(mgr, "deliver_queued_message", mock.MagicMock()),
            mock.patch.object(
                mgr.outbound_event, "wait", mock.CoroutineMock()
            ) as mock_wait,
            mock.patch.object(test_module, "trace_event", mock.MagicMock()),
        ):
            mock_wait.side_effect = KeyError()  # cover state=DELIVER logic and bail

            with self.assertRaises(KeyError):
                await mgr._process_loop()

    async def test_process_loop_x(self):
        mock_queued = mock.MagicMock(
            state=QueuedOutboundMessage.STATE_DONE,
            error=KeyError(),
            endpoint="http://1.2.3.4:8081",
            payload="Hello world",
        )

        self.profile = await create_test_profile()
        mock_handle_not_delivered = mock.MagicMock()
        mgr = OutboundTransportManager(self.profile, mock_handle_not_delivered)
        mgr.outbound_buffer.append(mock_queued)

        await mgr._process_loop()

    async def test_finished_deliver_x_log_debug(self):
        mock_queued = mock.MagicMock(state=QueuedOutboundMessage.STATE_DONE, retries=1)
        mock_completed_x = mock.MagicMock(exc_info=KeyError("an error occurred"))

        self.profile = await create_test_profile()
        mock_handle_not_delivered = mock.MagicMock()
        mgr = OutboundTransportManager(self.profile, mock_handle_not_delivered)
        mgr.outbound_buffer.append(mock_queued)
        with (
            mock.patch.object(test_module.LOGGER, "exception", mock.MagicMock()),
            mock.patch.object(test_module.LOGGER, "error", mock.MagicMock()),
            mock.patch.object(
                test_module.LOGGER, "isEnabledFor", mock.MagicMock()
            ) as mock_logger_enabled,
            mock.patch.object(mgr, "process_queued", mock.MagicMock()),
        ):
            mock_logger_enabled.return_value = True  # cover debug logging
            mgr.finished_deliver(mock_queued, mock_completed_x)

    async def test_should_encode_outbound_message(self):
        base_wire_format = BaseWireFormat()
        encoded_msg = "encoded_message"
        base_wire_format.encode_message = mock.CoroutineMock(return_value=encoded_msg)
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(BaseWireFormat, base_wire_format)
        self.profile.session = mock.CoroutineMock(return_value=mock.MagicMock())
        outbound = mock.MagicMock(payload="payload", enc_payload=None)
        target = mock.MagicMock()

        mgr = OutboundTransportManager(self.profile)
        result = await mgr.encode_outbound_message(self.profile, outbound, target)

        assert result.payload == encoded_msg
        base_wire_format.encode_message.assert_called_once_with(
            await self.profile.session(),
            outbound.payload,
            target.recipient_keys,
            target.routing_keys,
            target.sender_key,
        )

    async def test_should_not_encode_already_packed_message(self):
        self.profile = await create_test_profile()
        enc_payload = "enc_payload"
        outbound = mock.MagicMock(enc_payload=enc_payload)
        target = mock.MagicMock()

        mgr = OutboundTransportManager(self.profile)
        result = await mgr.encode_outbound_message(self.profile, outbound, target)

        assert result.payload == enc_payload



"""Schema artifacts."""

from marshmallow import Schema, fields

from ...messaging.models.openapi import OpenAPISchema
from ...messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
)


class CredDefValuePrimarySchema(OpenAPISchema):
    """Cred def value primary schema."""

    n = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )
    s = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )
    r = fields.Nested(
        Schema.from_dict(
            {
                "master_secret": fields.Str(
                    validate=NUM_STR_WHOLE_VALIDATE,
                    metadata={"example": NUM_STR_WHOLE_EXAMPLE},
                ),
                "number": fields.Str(
                    validate=NUM_STR_WHOLE_VALIDATE,
                    metadata={"example": NUM_STR_WHOLE_EXAMPLE},
                ),
                "remainder": fields.Str(
                    validate=NUM_STR_WHOLE_VALIDATE,
                    metadata={"example": NUM_STR_WHOLE_EXAMPLE},
                ),
            }
        ),
        metadata={"name": "CredDefValuePrimaryRSchema"},
    )
    rctxt = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )
    z = fields.Str(
        validate=NUM_STR_WHOLE_VALIDATE, metadata={"example": NUM_STR_WHOLE_EXAMPLE}
    )


class CredDefValueRevocationSchema(OpenAPISchema):
    """Cred def value revocation schema."""

    g = fields.Str(metadata={"example": "1 1F14F&ECB578F 2 095E45DDF417D"})
    g_dash = fields.Str(
        metadata={"example": "1 1D64716fCDC00C 1 0C781960FA66E3D3 2 095E45DDF417D"}
    )
    h = fields.Str(metadata={"example": "1 16675DAE54BFAE8 2 095E45DD417D"})
    h0 = fields.Str(metadata={"example": "1 21E5EF9476EAF18 2 095E45DDF417D"})
    h1 = fields.Str(metadata={"example": "1 236D1D99236090 2 095E45DDF417D"})
    h2 = fields.Str(metadata={"example": "1 1C3AE8D1F1E277 2 095E45DDF417D"})
    htilde = fields.Str(metadata={"example": "1 1D8549E8C0F8 2 095E45DDF417D"})
    h_cap = fields.Str(
        metadata={"example": "1 1B2A32CF3167 1 2490FEBF6EE55 1 0000000000000000"}
    )
    u = fields.Str(
        metadata={"example": "1 0C430AAB2B4710 1 1CB3A0932EE7E 1 0000000000000000"}
    )
    pk = fields.Str(
        metadata={"example": "1 142CD5E5A7DC 1 153885BD903312 2 095E45DDF417D"}
    )
    y = fields.Str(
        metadata={"example": "1 153558BD903312 2 095E45DDF417D 1 0000000000000000"}
    )


class CredDefValueSchema(OpenAPISchema):
    """Cred def value schema."""

    primary = fields.Nested(
        CredDefValuePrimarySchema(),
        metadata={"description": "Primary value for credential definition"},
    )
    revocation = fields.Nested(
        CredDefValueRevocationSchema(),
        metadata={"description": "Revocation value for credential definition"},
    )


class CredentialDefinitionSchema(OpenAPISchema):
    """Marshmallow schema for indy cred def."""

    ver = fields.Str(
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Node protocol version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    ident = fields.Str(
        data_key="id",
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    schemaId = fields.Str(
        metadata={
            "description": "Schema identifier within credential definition identifier",
            "example": ":".join(INDY_CRED_DEF_ID_EXAMPLE.split(":")[3:-1]),
        }
    )
    typ = fields.Constant(
        constant="CL",
        data_key="type",
        metadata={
            "description": "Signature type: CL for Camenisch-Lysyanskaya",
            "example": "CL",
        },
    )
    tag = fields.Str(
        metadata={
            "description": "Tag within credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE.split(":")[-1],
        }
    )
    value = fields.Nested(
        CredDefValueSchema(),
        metadata={"description": "Credential definition primary and revocation values"},
    )



from unittest import IsolatedAsyncioTestCase

from ...askar.profile import AskarProfile, AskarProfileSession
from ...config.injection_context import InjectionContext
from ...core.event_bus import EventBus
from ...core.protocol_registry import ProtocolRegistry
from ...protocols.coordinate_mediation.v1_0.route_manager import RouteManager
from ...resolver.base import BaseDIDResolver
from ...resolver.did_resolver import DIDResolver
from ...tests.mock import AsyncMock, MagicMock
from ...utils.stats import Collector
from ...utils.testing import create_test_profile
from ..adapters import ResolverAdapter, SecretsAdapter, SecretsAdapterError


class MockDIDResolver(BaseDIDResolver):
    async def setup(self, context: InjectionContext):
        return await super().setup(context)

    async def _resolve(self, profile, did):
        return await self.resolve(profile, did)

    async def resolve(
        self,
        profile,
        did,
        service_accept=None,
    ):
        return {"did": did, "test": "didDoc"}

    async def supports(self, profile, did: str):
        if did.startswith("did:test"):
            return True
        return False


class TestAdapters(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(settings={"experiment.didcommv2": True})
        self.profile.context.injector.bind_instance(ProtocolRegistry, ProtocolRegistry())
        self.profile.context.injector.bind_instance(Collector, Collector())
        self.profile.context.injector.bind_instance(EventBus, EventBus())
        self.profile.context.injector.bind_instance(RouteManager, MagicMock())
        self.test_did = "did:test:0"
        self.invalid_did = "this shouldn't work"
        resolver = DIDResolver()
        resolver.register_resolver(MockDIDResolver())
        self.res_adapter = ResolverAdapter(profile=self.profile, resolver=resolver)

    async def test_resolver_adapter_resolve_did(self):
        doc = await self.res_adapter.resolve(self.test_did)
        assert doc["did"] == self.test_did

    async def test_resolver_adapter_is_resolvable(self):
        valid = await self.res_adapter.is_resolvable(self.test_did)
        assert valid

        invalid = await self.res_adapter.is_resolvable(self.invalid_did)
        assert not invalid

    async def test_secrets_adapter_errors(self):
        sec_adapter = SecretsAdapter(session=MagicMock())
        with self.assertRaises(SecretsAdapterError) as ctx:
            await sec_adapter.get_secret_by_kid("kid")

        assert "ACA-Py's implementation of DMP only supports an Askar backend" in str(
            ctx.exception
        )

        store = MagicMock()
        askar_profile = AskarProfile(opened=store)
        session: AskarProfileSession = askar_profile.session()
        sec_adapter = SecretsAdapter(session)

        session._handle = MagicMock()
        session._handle.fetch_all_keys = AsyncMock(return_value=["key1", "key2"])

        with self.assertRaises(SecretsAdapterError) as ctx:
            await sec_adapter.get_secret_by_kid("kid")

        assert "More than one key found with kid" in str(ctx.exception)

    async def test_secrets_adapter_empty(self):
        store = MagicMock()
        askar_profile = AskarProfile(opened=store)
        session: AskarProfileSession = askar_profile.session()
        sec_adapter = SecretsAdapter(session)

        session._handle = MagicMock()
        session._handle.fetch_all_keys = AsyncMock(return_value=[None])

        assert not await sec_adapter.get_secret_by_kid("kid")

    async def test_secrets_adapter_valid_return(self):
        store = MagicMock()
        askar_profile = AskarProfile(opened=store)
        session: AskarProfileSession = askar_profile.session()
        sec_adapter = SecretsAdapter(session)

        entry = MagicMock()
        entry.key = "key1"

        session._handle = MagicMock()
        session._handle.fetch_all_keys = AsyncMock(return_value=[entry])

        key = await sec_adapter.get_secret_by_kid("kid")

        assert key.key == "key1"



from ...ld_proofs import DocumentVerificationResult, ProofResult, PurposeResult

CREDENTIAL_V2_TEMPLATE = {
    "@context": [
        "https://www.w3.org/ns/credentials/v2",
        "https://w3id.org/security/suites/ed25519-2020/v1",
    ],
    "type": ["VerifiableCredential"],
    "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
    "credentialSubject": {
        "id": "did:example:alice",
        "name": "Alice",
    },
}

CREDENTIAL_V2_ISSUED = {
    "@context": [
        "https://www.w3.org/ns/credentials/v2",
        "https://w3id.org/security/suites/ed25519-2020/v1",
    ],
    "type": ["VerifiableCredential"],
    "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
    "credentialSubject": {"id": "did:example:alice", "name": "Alice"},
    "proof": {
        "type": "Ed25519Signature2020",
        "proofPurpose": "assertionMethod",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "created": "2025-01-27T21:39:18+00:00",
        "proofValue": "zK9VFcysBRqQHQL65WNmKKPbYYrhFabu41SuQXMBGVEHHYLNGrELkNxg2GAxEs6phDZoGNcvhTBhv7fLmJ23U8Hn",
    },
}

CREDENTIAL_V2_VERIFIED = DocumentVerificationResult(
    verified=True,
    document={
        "@context": [
            "https://www.w3.org/ns/credentials/v2",
            "https://w3id.org/security/suites/ed25519-2020/v1",
        ],
        "type": ["VerifiableCredential"],
        "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
        "credentialSubject": {"id": "did:example:alice", "name": "Alice"},
        "proof": {
            "type": "Ed25519Signature2020",
            "proofPurpose": "assertionMethod",
            "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
            "created": "2025-01-27T21:39:18+00:00",
            "proofValue": "zK9VFcysBRqQHQL65WNmKKPbYYrhFabu41SuQXMBGVEHHYLNGrELkNxg2GAxEs6phDZoGNcvhTBhv7fLmJ23U8Hn",
        },
    },
    results=[
        ProofResult(
            verified=True,
            proof={
                "@context": [
                    "https://www.w3.org/ns/credentials/v2",
                    "https://w3id.org/security/suites/ed25519-2020/v1",
                ],
                "type": "Ed25519Signature2020",
                "proofPurpose": "assertionMethod",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2025-01-27T21:39:18+00:00",
                "proofValue": "zK9VFcysBRqQHQL65WNmKKPbYYrhFabu41SuQXMBGVEHHYLNGrELkNxg2GAxEs6phDZoGNcvhTBhv7fLmJ23U8Hn",
            },
            purpose_result=PurposeResult(
                valid=True,
                controller={
                    "@context": "https://w3id.org/security/v2",
                    "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                    "assertionMethod": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "authentication": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
                        }
                    ],
                    "capabilityDelegation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "capabilityInvocation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "keyAgreement": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6LSbkodSr6SU2trs8VUgnrnWtSm7BAPG245ggrBmSrxbv1R",
                            "type": "X25519KeyAgreementKey2019",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "5dTvYHaNaB7mk7iA9LqCJEHG2dGZQsvoi8WGzDRtYEf",
                        }
                    ],
                    "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                },
            ),
        )
    ],
)

PRESENTATION_V2_UNSIGNED = {
    "@context": ["https://www.w3.org/ns/credentials/v2"],
    "type": ["VerifiablePresentation"],
    "holder": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
    "verifiableCredential": [
        {
            "@context": [
                "https://www.w3.org/ns/credentials/v2",
                "https://w3id.org/security/suites/ed25519-2020/v1",
            ],
            "type": ["VerifiableCredential"],
            "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
            "credentialSubject": {"id": "did:example:alice", "name": "Alice"},
            "proof": {
                "type": "Ed25519Signature2020",
                "proofPurpose": "assertionMethod",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2025-01-27T21:39:18+00:00",
                "proofValue": "zK9VFcysBRqQHQL65WNmKKPbYYrhFabu41SuQXMBGVEHHYLNGrELkNxg2GAxEs6phDZoGNcvhTBhv7fLmJ23U8Hn",
            },
        }
    ],
}

PRESENTATION_V2_SIGNED = {
    "@context": [
        "https://www.w3.org/ns/credentials/v2",
        "https://w3id.org/security/suites/ed25519-2020/v1",
    ],
    "type": ["VerifiablePresentation"],
    "holder": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
    "verifiableCredential": [
        {
            "@context": [
                "https://www.w3.org/ns/credentials/v2",
                "https://w3id.org/security/suites/ed25519-2020/v1",
            ],
            "type": ["VerifiableCredential"],
            "issuer": {"id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"},
            "credentialSubject": {"id": "did:example:alice", "name": "Alice"},
            "proof": {
                "type": "Ed25519Signature2020",
                "proofPurpose": "assertionMethod",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2025-01-27T21:39:18+00:00",
                "proofValue": "zK9VFcysBRqQHQL65WNmKKPbYYrhFabu41SuQXMBGVEHHYLNGrELkNxg2GAxEs6phDZoGNcvhTBhv7fLmJ23U8Hn",
            },
        }
    ],
    "proof": {
        "type": "Ed25519Signature2020",
        "proofPurpose": "authentication",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "created": "2025-01-27T21:50:39+00:00",
        "challenge": "2b1bbff6-e608-4368-bf84-67471b27e41c",
        "proofValue": "z61aNLNSyBVZyYY5xEKYnGDzWbXQhpWa8QXmQMMJpy4zZ71kyxGbRHVwMWdEzU4qwQhLZ7eSfQiX4dENquYGxkbcB",
    },
}
